#include <iostream>
#include <memory>
using namespace std;

// Why and when would we need custom deleter?
// Case 1: In order to fully delete an object sometimes, we need to do some
//		additional action. What if performing "delete" (that smart pointers do
//		automatically) is not the only thing which needs to be done before fully
//		destroying the owned object.
// Case 2: We can’t bind a shared_ptr or unique_ptr to a stack-allocated object,
//		because calling delete on it would cause undefined behaviour.
// Case 3: Mix of programming languages code, such as C++ with Obj-C++.
//		As objective-c may need a complex release mechanism for its data types
//		such as calling CFRelease, we would be in need of a custom deleter.
// Case 4: In C where, when you wrap FILE*, or some kind of C style structure free(),
//		custom deleters may be useful.

struct SomeClass {
	SomeClass() { cout << "SomeClass(): " << this << endl; }
	~SomeClass() { cout << "~SomeClass()" << endl; }
};

struct Deleter {
	Deleter() {
		cout << "Deleter(): " << this << endl;
	}
	void operator()(SomeClass* p) const {
		cout << "Call delete for SomeClass object: " << p << endl;
		delete p;
	}
};

int main() {
	Deleter deleter;
	auto lam = [](auto *p) {
		cout << "Call lambda deleter for object: " << p << endl; delete p;
	};

	unique_ptr<SomeClass, decltype(lam)> u_ptr1(new SomeClass(), lam);
	unique_ptr<SomeClass, Deleter&> u_ptr2(new SomeClass(), deleter);
}

